package amiga;

import javax.imageio.ImageIO;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Properties;

/** The game main view class */
public final class TheAppAmiga extends JFrame implements
		WindowListener,
		KeyListener,
		ComponentListener,
		Runnable
{
	private static final long serialVersionUID = 1L;
	private static final String APP_NAME = "Stunt Car Racer";
	private static final String TEXT_ENCODE = "UTF-8";

	/** Refresh period in ms. 20 ms for PAL, 16.667 ms for NTSC */
	private static final long REFRESH_PERIOD_MS = 20;

	private static final int PICTURE_SCALE = 2;

	/** Screen width in px */
	private static final int SCREEN_WIDTH = 320 * PICTURE_SCALE;

	/** Screen height in px */
	private static final int SCREEN_HEIGHT = 200 * PICTURE_SCALE;

	private Game mGame = null;

	/** Top margin in px */
	private int mFrameViewportTopMargin = 0;

	/** Left margin in px */
	private int mFrameViewportLeftMargin = 0;

	/* Adopt to the current screen width size */
	// private double mPaintScaleWidth = 1.0;

	/* Adopt to the current screen height size */
	// private double mPaintScaleHeight = 1.0;

	/** Adopt to the current screen size keeping aspect ratio */
	private double mPaintScale = 1.0;

	/** A pointer to a current off-screen buffer to show */
	private Image mImgOffScreenBuffer = null;
	private Graphics mImgOffScreenBufferGraphics = null;

	/**
	 * Load an image with given name from image folder
	 *
	 * @param name an image name without extension. ".png" is used.
	 * @return the image or {@code null} if there is no image with the given
	 *         name.
	 */
	private static BufferedImage getImage(final String name) {
		try {
			return ImageIO.read( TheAppAmiga.class.getResource(
					"/images/" + name + ".png" ) );
		} catch( final Exception ie ) {
			return null;
		}
	}

	private TheAppAmiga() {
		setName( APP_NAME );
		setTitle( APP_NAME );
		setIconImage( getImage( "ico" ) );
		setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
		// Toolkit.getDefaultToolkit().setDynamicLayout( false );
		addWindowListener( this );
		setIgnoreRepaint( true );
		int width = SCREEN_WIDTH;
		int height = SCREEN_HEIGHT;
		final Rectangle r = getGraphicsConfiguration().getBounds();
		while( (width > r.width) || (height > r.height) ) {
			width >>= 1;
			height >>= 1;
		}
		getContentPane().setPreferredSize( new Dimension( width, height ) );
		pack();
		// final Insets in = getInsets();
		// setSize( width + in.left + in.right, height + in.top + in.bottom );
		// setSize( width, height );
		setLocationRelativeTo( null );// center the window on screen
		setResizable( true );
		setFocusable( true );
		addKeyListener( this );
		addComponentListener( this );

		setVisible( true );

		setLanguage( "en" );

		mGame = new Game();

		mImgOffScreenBuffer = new BufferedImage( Game.SCREEN_WIDTH, Game.SCREEN_HEIGHT, BufferedImage.TYPE_INT_ARGB );
		mImgOffScreenBufferGraphics = mImgOffScreenBuffer.getGraphics();

		new Thread( this, APP_NAME ).start();
		new Thread( mGame, APP_NAME ).start();
	}

	/**
	 * Internal dialog localization.
	 *
	 * @param name     resource name.
	 * @param language name of a language.
	 */
	private static void loadUIManagerLocalization(final String name, final String language) {
		InputStreamReader ir = null;
		try {
			final InputStream is = TheAppAmiga.class.getResourceAsStream(
					name + "_" + language + ".txt" );
			if( is == null ) {
				return;
			}
			ir = new InputStreamReader( is, TEXT_ENCODE );
			final Properties p = new Properties();
			p.load( ir );
			for( final Object k : p.keySet() ) {
				UIManager.put( k, p.getProperty( (String)k ) );
			}
		} catch( final Exception ignored) {
		} finally {
			if( ir != null ) {
				try { ir.close(); } catch( final IOException ignored) {}
			}
		}
	}

	private void setLanguage(final String language) {
		loadUIManagerLocalization( "FileChooser", language );
	}

	@Override
	public final void paint(final Graphics g) {
		synchronized( Game.class ) {
			final Graphics2D g2d = (Graphics2D) g;
			g2d.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF );
			// g2d.setRenderingHint( RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY );
			// g2d.setRenderingHint( RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_DEFAULT );
			g2d.setRenderingHint( RenderingHints.KEY_DITHERING, RenderingHints.VALUE_DITHER_DISABLE );
			g2d.setRenderingHint( RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON );
			// BICUBIC is the slowest, 16px, BILINEAR uses 4 px. NEAREST_NEIGHBOR is the fastest
			g2d.setRenderingHint( RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC );
			// g2d.setRenderingHint( RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_DEFAULT );
			// g2d.setRenderingHint( RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_DEFAULT );
			// g2d.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_DEFAULT );
			// g2d.setRenderingHint( RenderingHints.KEY_TEXT_LCD_CONTRAST, 255 );
			// g2d.scale( mPaintScaleWidth, mPaintScaleHeight );
			g2d.scale( mPaintScale, mPaintScale );
			g2d.drawImage( mImgOffScreenBuffer, mFrameViewportLeftMargin, mFrameViewportTopMargin, null );
			g2d.dispose();
		}
	}

	private void stop() {

	}

	// start KeyListener
	@Override
	public final void keyTyped(final KeyEvent e) {
		if( e.getKeyChar() != 0 ) {// 0 is set after reading
			mGame.setKeyTyped( e.getKeyChar() );
		}
	}

	@Override
	public final void keyPressed(final KeyEvent e) {
		mGame.setKey( e.getKeyCode() );
	}

	@Override
	public final void keyReleased(final KeyEvent e) {
		mGame.resetKey( e.getKeyCode() );
	}
	// end KeyListener

	// start ComponentListener
	@Override
	public final void componentResized(final ComponentEvent e) {
		final Insets inset = getInsets();
		/*
		mPaintScaleWidth =
				((double)(getWidth() - inset.left - inset.right) / (double)Game.SCREEN_WIDTH);
		mPaintScaleHeight =
				((double)(getHeight() - inset.top - inset.bottom) / (double)Game.SCREEN_HEIGHT);
		mFrameViewportTopMargin = (int) (inset.top / mPaintScaleHeight);
		mFrameViewportLeftMargin = (int) (inset.left / mPaintScaleWidth);
		*/
		int width = getWidth() - inset.left - inset.right;
		int height = getHeight() - inset.top - inset.bottom;
		if( height > width * Game.SCREEN_HEIGHT / Game.SCREEN_WIDTH ) {
			// vertical margins
			mPaintScale = (float) width / Game.SCREEN_WIDTH;
			mFrameViewportTopMargin =
					(int)(inset.top + (0.5 * ((double) height - (double) width * Game.SCREEN_HEIGHT / Game.SCREEN_WIDTH)) / mPaintScale);
			mFrameViewportLeftMargin = (int) (inset.left / mPaintScale);
		} else {
			// horizontal margins
			mPaintScale = ((double) height / (double) Game.SCREEN_HEIGHT);
			mFrameViewportTopMargin = (int) (inset.top / mPaintScale);
			mFrameViewportLeftMargin =
					(int) (inset.left + (0.5 * ((double) width - ((double) height * Game.SCREEN_WIDTH / Game.SCREEN_HEIGHT))) / mPaintScale);
		}
	}

	@Override
	public final void componentMoved(final ComponentEvent e) {
	}

	@Override
	public void componentShown(final ComponentEvent e) {
	}

	@Override
	public void componentHidden(final ComponentEvent e) {
	}
	// end ComponentListener

	// start WindowListener
	@Override
	public final void windowOpened(final WindowEvent e) {
	}

	@Override
	public final void windowClosing(final WindowEvent e) {
		stop();
		setVisible( false );
		dispose();
		System.exit( 0 );
	}

	@Override
	public final void windowClosed(final WindowEvent e) {
	}

	@Override
	public final void windowIconified(final WindowEvent e) {
	}

	@Override
	public final void windowDeiconified(final WindowEvent e) {
	}

	@Override
	public final void windowActivated(final WindowEvent e) {
	}

	@Override
	public final void windowDeactivated(final WindowEvent e) {
	}
	// end WindowListener

	private void waiting(long time) {
		try {
			synchronized( Game.class ) {
				final Graphics g = mImgOffScreenBufferGraphics;
				g.drawImage( mGame.getScene(), 0, 0, null );
				Game.vertbInterrupt( g );
				repaint();
			}
			time += REFRESH_PERIOD_MS;
			time -= System.currentTimeMillis();
			if( time > 0 ) {
				Thread.sleep( time );
			}
		} catch(final Exception ignored) {}
	}

	@Override
	public final void run() {

		while( true ) {
			final long curr_time = System.currentTimeMillis();
			waiting( curr_time );
		} // while running
		/* while( true ) {
			try {
				mThreadGame.join();
				if( mGame != null ) {
					mGame.release();
					mGame = null;
				}
				return;
			} catch(final InterruptedException e) {// ignored
			}
		} */
	}

	private static void log(final Throwable t) {
		System.err.println( t.getMessage() );
		t.printStackTrace();
		/*
		final String newline = "\n";
		final JTextArea message = new JTextArea( t.toString() );
		message.append( newline );
		message.setFont( new Font( "Dialog", Font.PLAIN, 12 ) );
		final StackTraceElement[] stack = t.getStackTrace();
		for( final StackTraceElement s : stack ) {
			message.append("    ");
			message.append( s.toString() );
			message.append( newline );
		}
		final JScrollPane scroll = new JScrollPane( message );
		scroll.setPreferredSize( new Dimension( 640, 320 ) );
		final JPanel p = new JPanel( new BorderLayout() );
		p.add( new JLabel( Localization.get("Log.MsgError"), JLabel.CENTER ), BorderLayout.PAGE_START );
		p.add( scroll, BorderLayout.CENTER );
		if( JOptionPane.NO_OPTION ==
				JOptionPane.showConfirmDialog( null, p, Localization.get("Log.MsgErrorTitle"),
						JOptionPane.ERROR_MESSAGE, JOptionPane.YES_NO_OPTION ) ) {
			System.exit( 1 );
		}
		*/
	}

	/**
	 * Entry point
	 *
	 * @param args is not used
	 */
	public static void main(final String[] args) {

		if( GraphicsEnvironment.isHeadless() ) {
			return;
		}

		SwingUtilities.invokeLater( new Runnable() {

			@Override
			public final void run() {
				// Thread.currentThread().setName( APP_NAME );
				Thread.setDefaultUncaughtExceptionHandler( new Thread.UncaughtExceptionHandler() {
					@Override
					public final void uncaughtException(final Thread t, final Throwable e) {
						TheAppAmiga.log( e );
					}
				} );
				new TheAppAmiga();
			}
		} );
	}
}
